import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import org.encog.engine.network.activation.ActivationTANH;
import org.encog.mathutil.randomize.NguyenWidrowRandomizer;
import org.encog.neural.data.NeuralData;
import org.encog.neural.data.NeuralDataSet;
import org.encog.neural.data.basic.BasicNeuralData;
import org.encog.neural.data.basic.BasicNeuralDataSet;
import org.encog.neural.networks.BasicNetwork;
import org.encog.neural.networks.layers.BasicLayer;
import org.encog.neural.networks.logic.FeedforwardLogic;
import org.encog.neural.networks.training.Train;
import org.encog.neural.networks.training.propagation.resilient.ResilientPropagation;
import org.encog.util.logging.Logging;

/**
 * @author Dominika Fedakova
 *
 * Trieda create_NN ponuka niekolko funkcii na vytvorenie
 * a pracu s doprednou neuronovou sietou a uciacim 
 * algoritmom resilient propagation. Ponuka funkcie na
 * citanie a zapisovanie do cvs suboru, pre lahke 
 * importovanie dat do siete.
 */

public class create_NN {

	/**
	    * Class Normalizuje realnu hodnotu do intervalu <-1,1>
	    *
	    * @param double Konkretne cislo na normalizaciu
	    * @param double Minimalna hodnota intervalu, z ktoreho je cislo
	    * @param double Maximalna hodnota intervalu, z ktoreho je cislo
	    * @return double Normalizovana hodnota cisla
	    */
	public static double normalize(double number,double d_min, double d_max){
		//funkcia vracajuca normalizovanu hodnotu number, dolezity je aj interval, z ktoreho je number
		//pricom interval urcuju hodnoty d_min a d_max
		double norm_number;
		norm_number=(number-d_min)*2/(d_max-d_min)-1;
		return norm_number;
	}
	
	/**
	    * Class Denormalizuje hodnotu z intervalu <-1,1> na realnu hodnotu
	    *
	    * @param double Konkretne cislo na denormalizaciu
	    * @param double Minimalna hodnota intervalu, do ktoreho chceme denormalizovat
	    * @param double Maximalna hodnota intervalu, do ktoreho chceme denormalizovat
	    * @return double Denormalizovana hodnota cisla
	    */
	public static double denormalize(double number,double d_min, double d_max){
		//opacna funkcia k normalizacii, teda ak chceme denormalizovat data
		double denorm_number;
		denorm_number=(((d_max-d_min)*(number+1))/2)+d_min;
		return denorm_number;
	}
	
	/**
	    * pocet vzoriek v datach nacitanych zo suboru
	*/
	public static int num_data=0;
	/**
	    * pocet vstupnych premennych
	*/
	public static int num_inputs=0;
	public static int size=3000;
	/**
	    * chyba MSE natrenovanej siete
	*/
	public static double error=0;
	/**
	    * pole maximalnych hodnot jednotlivych vstupnych premennych
	*/
	public static double[] data_max;
	/**
	    * pole minimalnych hodnot jednotlivych vstupnych premennych
	*/
	public static double[] data_min;
	/**
	    * pole vstupnych premennych
	*/
	public static double[][] input;
	/**
	    * pole vystupnych premennych
	*/
	public static double[][] output;
	
	
	/**
	    * Class Nacitava data zo siete a uklada vstupne premenne
	    * do pola input a vystupne do pola output
	    * @param File Subor, v ktorom mame ulozene data na trenovanie siete
	    */
	public static void read_csv_file(File file) throws IOException {
		//nacitanie dat
		double[][] in_array=new double[size][];
		double[][] out_array=new double[size][1];
		String[] variables;
		num_data=0;
		num_inputs=0;
		BufferedReader br=new BufferedReader(new FileReader(file));
		String line=br.readLine();
		if(line.contains(";")){
				if (line.contains(",")){
					line=line.replace(",", ".");
				}
				variables=line.split(";");
			  } else {
				variables=line.split(",");
		}
		num_inputs=variables.length-1;
		data_max=new double[num_inputs+1];
		data_min=new double[num_inputs+1];
		for (int i=0;i<num_inputs+1;i++){
			data_max[i]=0;
			data_min[i]=0;
		} 
		do {
			if (line!=""){
			  if(line.contains(";")){
				if (line.contains(",")){
					line=line.replace(",", ".");
				}
				variables=line.split(";");
			  } else {
				variables=line.split(",");
			}
			in_array[num_data]=new double[num_inputs];
			for (int i=0;i<num_inputs;i++){
				in_array[num_data][i]=Double.parseDouble(variables[i]);
				//upravime este maxima a minima celych dat ak je to potrebne
				if (in_array[num_data][i]>data_max[i]){
					data_max[i]=in_array[num_data][i];
				}
				if (in_array[num_data][i]<data_min[i]){
					data_min[i]=in_array[num_data][i];
				}
				
			}
			out_array[num_data][0]=Double.parseDouble(variables[num_inputs]);
			//upravime este maxima a minima celych dat ak je to potrebne
			if (out_array[num_data][0]>data_max[num_inputs]){
				data_max[num_inputs]=out_array[num_data][0];
			}
			if (out_array[num_data][0]<data_min[num_inputs]){
				data_min[num_inputs]=out_array[num_data][0];
			}
			num_data++;
			//ak nemame dostatocne velke pole vytvorime vacsie nove pole a prekopirujeme stare do noveho
			if (num_data==size){
				size=size*2;
				double[][] temp=new double[size][num_inputs];
				double[][] temp1=new double[size][1];
				System.arraycopy(in_array, 0, temp, 0, in_array.length);
				System.arraycopy(out_array, 0, temp1, 0, out_array.length);
				in_array=temp;
				out_array=temp1;
			}
			}
		} while ((line=br.readLine())!=null);
		//musime normalizovat hodnoty, a vytvorit staticke pole presnej velkosti ako su data
		input=new double[num_data][num_inputs];
		output=new double[num_data][1];
		for (int i=0;i<num_data;i++){
			for (int j=0;j<num_inputs;j++){
				input[i][j]=normalize(in_array[i][j],data_min[j],data_max[j]);
			}
			output[i][0]=normalize(out_array[i][0],data_min[num_inputs],data_max[num_inputs]);
		}
	}
	
	/**
	    * Class Funkcia vytvara novu doprednu siet s uciacim algoritmom resilient propagation
	    * @param File Subor, v ktorom mame ulozene trenovacie data
	    * @param int Maximalny pocet epochov pri trenovani siete
	    * @param double Minimalna chcena chyba MSE siete, podmienka ukoncenia trenovania siete
	    * @param int Pocet neuronov v skrytej vrstve, minimalne dvojnasobny pocet 
	    * vstupnych premennych( ak je pocet mensi, defaulte sa nastavi na dvojnasobok)
	    * @throws IOException Chyba pri citani pri subore 
	    */
	public static BasicNetwork create_RPROP(File file, int maxEpoch, double minError, int num_Hneuron) throws IOException {
		Logging.stopConsoleLogging();
		//testovanie dostatocneho poctu skrytych neuronov
		if (num_Hneuron<2*num_inputs){
			num_Hneuron=2*num_inputs;
		}
		//nacitame data
		read_csv_file(file);
		//vlozime data do tvaru neuronovej siete
		NeuralDataSet trainingSet = new BasicNeuralDataSet(input, output);
		//zalozime novu siet
		BasicNetwork network = new BasicNetwork();
		//pridame jej vrstvy
		network.addLayer(new BasicLayer(new ActivationTANH(),true,num_inputs));
		network.addLayer(new BasicLayer(new ActivationTANH(),true,num_Hneuron));
		network.addLayer(new BasicLayer(new ActivationTANH(),true,1));
		//nastavime, aby to bola dopredna siet
		network.setLogic(new FeedforwardLogic());
		//a mozeme ukonsit strukturu siete
		network.getStructure().finalizeStructure();
		//zinicializuje, zaciatocne vahy nastavime nahodne z intervalu <-1,1>
		(new NguyenWidrowRandomizer(-1,1)).randomize(network);
		//nastavime uciaci algoritmus resilient propagation
		final Train train = new ResilientPropagation(network, trainingSet);
		
		/*
		  strategia reset, po 50 iteraciach s nezmenenou chybou
		  train.addStrategy(new RequiredImprovementStrategy(50));
		*/
		
		//zacneme sa ucit po jednotlivych iteraciach
		/**
		    * pocet epochov pri iteraciach ucenia siete
		*/
		int epoch = 1;
		do {
			//priami prikaz na trenovanie siete v dalsej iteracii
			train.iteration();
			//vzdy vypiseme MSE siete, a sledujeme priebezny stav ucenia siete
			System.out.println("Epoch #" + epoch + " Error:" + denormalize(train.getError(),-data_max[num_inputs],data_max[num_inputs]));
			epoch++;
			//musime kontrolovat, aby sme nepresiahli maximalny pocet epochov
			if ((epoch>maxEpoch) && (maxEpoch>-1)) break;
			//podmienka skoncenia treningu je urcene chcene minimalne MSE
		} while(denormalize(train.getError(),-data_max[num_inputs],data_max[num_inputs]) >minError);
		
		//nakoniec este vypiseme koncovu MSE chybu siete
		System.out.println("Neural Network Results:");
		System.out.println(error=denormalize(train.getError(),-data_max[num_inputs],data_max[num_inputs]));
		
		//funkcia vracia natrenovanu siet
		return network;
	}
	
	/**
	    * Class Zaokruhluje hodnotu na dany pocet desatinych miest
	    * @param double Cislo na zaokruhlovanie
	    * @param int Pocet desatinych miest
	    */
	public static double round(double x, int pocet) {
		//funkcia na zaokruhlenie na pocet desatinnych miest
		return Math.round( x * Math.pow(10,pocet))/(double)Math.pow(10,pocet);
	}
	
	/**
	    * Class Funkcia rata vystup pri zadanych vstupnych premennych
	    * a natrenovanej sieti
	    * @param double[][] Pole vstupnych premennych
	    * @param BasicNetwork Natrenovana siet, ktora bude ratat vystup
	    * @return double Vypocitany denormalizovany vystup siete
	    */
	public static double recognize_output(double[] inputs,BasicNetwork network1 ) {
		double NN_output;double temp;
		for (int i=0; i<inputs.length;i++){
			temp=normalize(inputs[i],data_min[i],data_max[i]);
			inputs[i]=temp;
		}
		NeuralData input=new BasicNeuralData(inputs);
		NeuralData output=network1.compute(input);
		NN_output=denormalize(output.getData(0),data_min[num_inputs],data_max[num_inputs]);	
		return NN_output;
	}
	
	/**
	    * Class Funkcia rata vystupy pre sady zadanych vstupnych premennych
	    * ulozenych v subore cvs a zapisuje vstupy spolu s vystupmi do
	    * suboru cvs
	    * do pola input a vystupne do pola output
	    * @param File Subor, v ktorom mame ulozene data, sady vstupnych premennych
	    * na ratanie outputu (kazdy riadok obsahuje novu sadu premennych)
	    * @param String Nazov, pre subor so zapisanymi vyratami vystupnymi hodnotami
	    * @param BasicNatwork Natrenovana siet ratajuca vystupy
	    * @throws IOException Chyba pri citani pri subore 
	    */
	public static void recognize_file_outputs(File file,String file_out,BasicNetwork network1) throws IOException {
		String[] file_inputs;
		double[] double_inputs;
		BufferedReader br=new BufferedReader(new FileReader(file));
		BufferedWriter bw=new BufferedWriter(new FileWriter(file_out+".csv"));
		String line;
		while((line=br.readLine())!=null){
		if(line.contains(";")){
				if (line.contains(",")){
					line=line.replace(",", ".");
				}
				file_inputs=line.split(";");
				double_inputs=new double[file_inputs.length];
				for (int i=0;i<file_inputs.length;i++){
					double_inputs[i]=Double.parseDouble(file_inputs[i]);
				}
				bw.append(line+";"+recognize_output(double_inputs,network1)+"\n");
			  } else {
				file_inputs=line.split(",");
				double_inputs=new double[file_inputs.length];
				for (int i=0;i<file_inputs.length;i++){
					double_inputs[i]=Double.parseDouble(file_inputs[i]);
				}
				bw.append(line+","+recognize_output(double_inputs,network1)+"\n");
			  }
		}
		bw.close();
	}
	
}
